from dataclasses import dataclass
from enum import Enum
from typing import Optional


class WorkFlowType(Enum):
    GROW_OPTIMIZER = "grow_optimizer"
    GROW_SIMULATOR = "grow_simulator"


@dataclass
class PostgresConfig:
    host: str
    port: int
    database_name: str
    user_name: str
    password: str


@dataclass
class RabbitmqConfig:
    host: str
    port: int
    exchange_name: str
    user_name: str
    password: str
    hipe_compile: Optional[int] = 1



import os

from nwnsdk.app_logging import setup_logging, LogLevel

setup_logging(LogLevel.parse(os.environ.get("LOG_LEVEL", "INFO")), "nwnsdk")

from nwnsdk.config import WorkFlowType, PostgresConfig, RabbitmqConfig
from nwnsdk.postgres.dbmodels import JobStatus
from nwnsdk.rabbitmq.rabbitmq_client import Queue
from nwnsdk.nwn_client import NwnClient



from dataclasses import dataclass
from typing import List, Callable, Dict
from uuid import uuid4

from nwnsdk.postgres.dbmodels import Job
from nwnsdk.rabbitmq.rabbitmq_client import RabbitmqClient, Queue, PikaCallback

import logging
from nwnsdk.postgres.postgres_client import PostgresClient
from nwnsdk import PostgresConfig, RabbitmqConfig, WorkFlowType, JobStatus

LOGGER = logging.getLogger("nwnsdk")


class NwnClient(PostgresClient, RabbitmqClient):
    rabbitmq_client: RabbitmqClient
    postgres_client: PostgresClient
    logger: logging.Logger

    def __init__(self, postgres_config: PostgresConfig, rabbitmq_config: RabbitmqConfig):
        PostgresClient.__init__(self, postgres_config)
        RabbitmqClient.__init__(self, rabbitmq_config)

    def connect(self):
        PostgresClient._connect_postgres(self)
        RabbitmqClient._start_rabbitmq(self)

    def stop(self):
        PostgresClient._close_postgres(self)
        RabbitmqClient._stop_rabbitmq(self)

    def start_work_flow(
        self, work_flow_type: WorkFlowType, job_name: str, esdl_str: str, user_name: str, project_name: str
    ) -> uuid4:
        job_id: uuid4 = uuid4()
        PostgresClient._send_input(
            self,
            job_id=job_id,
            job_name=job_name,
            work_flow_type=work_flow_type,
            esdl_str=esdl_str,
            user_name=user_name,
            project_name=project_name,
        )
        RabbitmqClient._send_start_work_flow(self, job_id, work_flow_type)

        return job_id

    def get_job_details(self, job_id: uuid4) -> Job:
        return self.get_job(job_id)

    def get_all_jobs(self) -> List[Job]:
        return self.get_jobs()

    def get_jobs_from_ids(self, job_ids: List[uuid4]) -> List[Job]:
        return self.get_jobs(job_ids)



from dataclasses import dataclass
from enum import Enum

import sqlalchemy as db
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.dialects.postgresql import UUID

from nwnsdk import WorkFlowType

Base = declarative_base()


class JobStatus(Enum):
    REGISTERED = "registered"
    RUNNING = "running"
    FINISHED = "finished"
    ERROR = "error"
    STOPPED = "stopped"


@dataclass
class Job(Base):
    __tablename__ = "job"

    job_id = db.Column(UUID(as_uuid=True), primary_key=True)
    job_name = db.Column(db.String, nullable=False)
    work_flow_type = db.Column(db.Enum(WorkFlowType), nullable=False)
    user_name = db.Column(db.String)
    project_name = db.Column(db.String)
    status = db.Column(db.Enum(JobStatus), nullable=False)
    input_config = db.Column(db.String)
    input_esdl = db.Column(db.String, nullable=False)
    output_esdl = db.Column(db.String)
    added_at = db.Column(db.DateTime(timezone=True), nullable=False)
    running_at = db.Column(db.DateTime(timezone=True))
    stopped_at = db.Column(db.DateTime(timezone=True))
    logs = db.Column(db.String)



#  Copyright (c) 2023. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Setup logging for this python application."""

import logging
import sys
from enum import Enum
from typing import Optional, Dict

CONFIGURED_LOGGERS: Dict[str, logging.Logger] = {}


class LogLevel(Enum):
    """Simple enum to cover log levels for logging library."""

    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING
    ERROR = logging.ERROR

    @staticmethod
    def parse(value: str) -> "LogLevel":
        """
        Parses a given string for LogLevel's.

        Parameters
        ----------
        value : str
                user provided string containing the requested log level

        Returns
        -------
        LogLevel
            Loglevel for this logger
        """
        lowered = value.lower()

        if lowered == "debug":
            result = LogLevel.DEBUG
        elif lowered == "info":
            result = LogLevel.INFO
        elif lowered in ["warning", "warn"]:
            result = LogLevel.WARNING
        elif lowered in ["err", "error"]:
            result = LogLevel.ERROR
        else:
            raise ValueError(f"Value {value} is not a valid log level.")

        return result


def setup_logging(log_level: LogLevel, logger_name: Optional[str]) -> logging.Logger:
    """
    Initializes logging.

    Parameters
    ----------
    log_level : LogLevel
        The LogLevel for this logger.
    logger_name : Optional[str]
        Name for this logger.
    """
    logger = logging.getLogger(logger_name)

    if logger_name not in CONFIGURED_LOGGERS:
        print(f"Will use log level {log_level} for logger {logger_name}")
        logger.setLevel(log_level.value)

        log_handler = logging.StreamHandler(sys.stdout)
        formatter = logging.Formatter(
            fmt="%(asctime)s [%(threadName)s][%(filename)s:%(lineno)d]" "[%(levelname)s]: %(message)s"
        )
        log_handler.setFormatter(formatter)
        logger.addHandler(log_handler)
        CONFIGURED_LOGGERS[logger_name] = logger

    return logger



from datetime import datetime
import logging
from typing import List
from uuid import uuid4

from sqlalchemy import select, update, delete
from sqlalchemy.orm import Session
from sqlalchemy.orm.strategy_options import load_only
from sqlalchemy.engine import Engine

from nwnsdk import PostgresConfig, WorkFlowType
from nwnsdk.postgres.database import initialize_db, session_scope
from nwnsdk.postgres.dbmodels import Job, JobStatus

LOGGER = logging.getLogger("nwnsdk")

ALL_JOBS_STMNT = select(Job).options(
    load_only(
        Job.job_id,
        Job.job_name,
        Job.work_flow_type,
        Job.user_name,
        Job.project_name,
        Job.status,
        Job.added_at,
        Job.running_at,
        Job.stopped_at,
    )
)


class PostgresClient:
    db_config: PostgresConfig
    engine: Engine

    def __init__(self, postgres_config: PostgresConfig):
        self.db_config = postgres_config

    def _connect_postgres(self):
        self.engine = initialize_db("nwn", self.db_config)

    def _close_postgres(self):
        if self.engine:
            self.engine.dispose()

    def _send_input(
        self,
        job_id: uuid4,
        job_name: str,
        work_flow_type: WorkFlowType,
        esdl_str: str,
        user_name: str,
        project_name: str,
    ) -> None:
        with session_scope() as session:
            new_job = Job(
                job_id=job_id,
                job_name=job_name,
                work_flow_type=work_flow_type,
                user_name=user_name,
                project_name=project_name,
                status=JobStatus.REGISTERED,
                input_esdl=esdl_str,
                added_at=datetime.now(),
            )
            session.add(new_job)

    def set_job_running(self, job_id: uuid4) -> None:
        LOGGER.debug("Started job with id '%s'", job_id)
        with session_scope() as session:
            stmnt = update(Job).where(Job.job_id == job_id).values(status=JobStatus.RUNNING, running_at=datetime.now())
            session.execute(stmnt)

    def store_job_result(self, job_id: uuid4, new_logs: str, new_status: JobStatus, output_esdl: str):
        LOGGER.debug(
            "Stored job result %s with exit code %s, status %s and %s characters of log",
            job_id,
            new_status,
            len(new_logs),
        )
        with session_scope() as session:
            stmnt = (
                update(Job)
                .where(Job.job_id == job_id)
                .values(status=new_status, logs=new_logs, output_esdl=output_esdl, stopped_at=datetime.now())
            )
            session.execute(stmnt)

    def get_job_status(self, job_id: uuid4) -> JobStatus:
        LOGGER.debug("Retrieving job status for job with id '%s'", job_id)
        with session_scope(do_expunge=True) as session:
            stmnt = select(Job.status).where(Job.job_id == job_id)
            job_status = session.scalar(stmnt)
        return job_status

    def get_job_input_esdl(self, job_id: uuid4) -> str:
        LOGGER.debug("Retrieving input esdl for job %s", job_id)
        with session_scope(do_expunge=True) as session:
            stmnt = select(Job.input_esdl).where(Job.job_id == (job_id))
            job_input_esdl: str = session.scalar(stmnt)
        return job_input_esdl

    def get_job_output_esdl(self, job_id: uuid4) -> str:
        LOGGER.debug("Retrieving job output esdl for job with id '%s'", job_id)
        with session_scope() as session:
            stmnt = select(Job.output_esdl).where(Job.job_id == job_id)
            job_output_esdl: Job = session.scalar(stmnt)
        return job_output_esdl

    def get_job_logs(self, job_id: uuid4) -> str:
        LOGGER.debug("Retrieving job log for job with id '%s'", job_id)
        with session_scope() as session:
            stmnt = select(Job.logs).where(Job.job_id == job_id)
            job_logs: Job = session.scalar(stmnt)
        return job_logs

    def get_job(self, job_id: uuid4) -> Job:
        LOGGER.debug("Retrieving job data for job with id '%s'", job_id)
        session: Session
        with session_scope(do_expunge=True) as session:
            stmnt = select(Job).where(Job.job_id == job_id)
            job = session.scalar(stmnt)
        return job

    def get_jobs(self, job_ids: List[uuid4] = None) -> List[Job]:
        with session_scope(do_expunge=True) as session:
            stmnt = ALL_JOBS_STMNT
            if job_ids:
                LOGGER.debug(f"Retrieving job data for jobs '{','.join([str(job_id) for job_id in job_ids])}'")
                stmnt = stmnt.where(Job.job_id.in_(job_ids))
            else:
                LOGGER.debug(f"Retrieving job data for all jobs")

            jobs = session.scalars(stmnt).all()
        return jobs

    def get_jobs_from_user(self, user_name: str) -> List[Job]:
        LOGGER.debug(f"Retrieving job data for jobs from user '{user_name}'")
        with session_scope(do_expunge=True) as session:
            stmnt = ALL_JOBS_STMNT.where(Job.user_name == user_name)
            jobs = session.scalars(stmnt).all()
        return jobs

    def get_jobs_from_project(self, project_name: str) -> List[Job]:
        LOGGER.debug(f"Retrieving job data for jobs from project '{project_name}'")
        with session_scope(do_expunge=True) as session:
            stmnt = ALL_JOBS_STMNT.where(Job.project_name == project_name)
            jobs = session.scalars(stmnt).all()
        return jobs

    def delete_job(self, job_id: uuid4) -> bool:
        LOGGER.debug("Deleting job with id '%s'", job_id)
        session: Session
        with session_scope() as session:
            stmnt = select(Job).where(Job.job_id == job_id)
            job = session.scalars(stmnt).all()
            if job:
                stmnt = delete(Job).where(Job.job_id == job_id)
                session.execute(stmnt)
                job_deleted = True
            else:
                job_deleted = False
        return job_deleted



import logging
from contextlib import contextmanager
from typing import Generator

from sqlalchemy import create_engine, orm
from sqlalchemy.engine import URL
from sqlalchemy.orm import Session as SQLSession

from nwnsdk import PostgresConfig

LOGGER = logging.getLogger("nwnsdk")

session_factory = orm.sessionmaker()
Session = orm.scoped_session(session_factory)


@contextmanager
def session_scope(do_expunge=False) -> Generator[SQLSession, None, None]:
    """Provide a transactional scope around a series of operations. Ensures that the session is
    committed and closed. Exceptions raised within the 'with' block using this contextmanager
    should be handled in the with block itself. They will not be caught by the 'except' here."""
    try:
        yield Session()

        if do_expunge:
            Session.expunge_all()
        Session.commit()
    except Exception as e:
        # Only the exceptions raised by session.commit above are caught here
        Session.rollback()
        raise e
    finally:
        Session.remove()


def initialize_db(application_name: str, config: PostgresConfig):
    """
    Initialize the database connection by creating the engine and configuring
    the default session maker.
    """
    LOGGER.info("Connecting to PostgresDB at %s:%s as user %s", config.host, config.port, config.user_name)
    url = URL.create(
        "postgresql+psycopg2",
        username=config.user_name,
        password=config.password,
        host=config.host,
        port=config.port,
        database=config.database_name,
    )

    engine = create_engine(
        url,
        pool_size=20,
        max_overflow=5,
        echo=True,
        connect_args={
            "application_name": application_name,
            "options": "-c lock_timeout=30000 -c statement_timeout=300000",  # 5 minutes
        },
    )

    # Bind the global session to the actual engine.
    Session.configure(bind=engine)

    return engine



#!/usr/bin/env python
import functools
import logging
import threading
from enum import Enum
from typing import Callable, Dict, Optional
from uuid import uuid4

import pika
import pika.exceptions
import json

from pika.adapters.blocking_connection import BlockingChannel

from nwnsdk import RabbitmqConfig, WorkFlowType

LOGGER = logging.getLogger("nwnsdk")


PikaCallback = Callable[
    [pika.adapters.blocking_connection.BlockingChannel, pika.spec.Basic.Deliver, pika.spec.BasicProperties, bytes], None
]


class Queue(Enum):
    StartWorkflowOptimizer = "start_work_flow.optimizer"
    StartWorkflowGrowSimulator = "start_work_flow.grow_simulator"

    @staticmethod
    def from_workflow_type(workflow_type: WorkFlowType) -> "Queue":
        if workflow_type == WorkFlowType.GROW_OPTIMIZER:
            return Queue.StartWorkflowOptimizer
        elif workflow_type == WorkFlowType.GROW_SIMULATOR:
            return Queue.StartWorkflowGrowSimulator
        else:
            raise RuntimeError(f"Unimplemented workflow type {workflow_type}. Please implement.")


class RabbitmqClient(threading.Thread):
    rabbitmq_callbacks: Dict[Queue, PikaCallback]
    rabbitmq_is_running: bool
    rabbitmq_config: RabbitmqConfig
    rabbitmq_exchange: str
    rabbitmq_connection: Optional[pika.BlockingConnection]
    rabbitmq_channel: Optional[BlockingChannel]

    def __init__(self, config: RabbitmqConfig):
        super().__init__()
        self.rabbitmq_callbacks = {}
        self.rabbitmq_is_running = False
        self.rabbitmq_config = config
        self.rabbitmq_exchange = config.exchange_name
        self.rabbitmq_connection = None
        self.rabbitmq_channel = None

    def _connect_rabbitmq(self):
        # initialize rabbitmq connection
        LOGGER.info(
            "Connecting to RabbitMQ at %s:%s as user %s",
            self.rabbitmq_config.host,
            self.rabbitmq_config.port,
            self.rabbitmq_config.user_name,
        )
        credentials = pika.PlainCredentials(self.rabbitmq_config.user_name, self.rabbitmq_config.password)
        parameters = pika.ConnectionParameters(
            self.rabbitmq_config.host,
            self.rabbitmq_config.port,
            "/",
            credentials,
            heartbeat=60,
            blocked_connection_timeout=3600,
            connection_attempts=10,
        )

        if not self.rabbitmq_connection or self.rabbitmq_connection.is_closed:
            LOGGER.info("Setting up a new connection to RabbitMQ.")
            self.rabbitmq_connection = pika.BlockingConnection(parameters)

        if not self.rabbitmq_channel or self.rabbitmq_channel.is_closed:
            LOGGER.info("Setting up a new channel to RabbitMQ.")
            self.rabbitmq_channel = self.rabbitmq_connection.channel()
            self.rabbitmq_channel.basic_qos(prefetch_size=0, prefetch_count=1)
            self.rabbitmq_channel.exchange_declare(exchange=self.rabbitmq_exchange, exchange_type="topic")
            for queue_item in Queue:
                queue = self.rabbitmq_channel.queue_declare(queue_item.value, exclusive=False).method.queue
                self.rabbitmq_channel.queue_bind(queue, self.rabbitmq_exchange, routing_key=queue_item.value)

            for queue, callback in self.rabbitmq_callbacks.items():
                self.rabbitmq_channel.basic_consume(queue=queue.value, on_message_callback=callback, auto_ack=False)
        LOGGER.info("Connected to RabbitMQ")

    def _start_rabbitmq(self):
        self._connect_rabbitmq()
        self.start()

    def set_callbacks(self, callbacks: Dict[Queue, PikaCallback]):
        self.rabbitmq_callbacks.update(callbacks)
        for queue, callback in callbacks.items():
            self.rabbitmq_connection.add_callback_threadsafe(
                functools.partial(
                    self.rabbitmq_channel.basic_consume, queue=queue.value, on_message_callback=callback, auto_ack=False
                )
            )

    def run(self):
        self.rabbitmq_is_running = True

        while self.rabbitmq_is_running:
            try:
                LOGGER.info("Waiting for input...")
                while self.rabbitmq_is_running:
                    self.rabbitmq_channel._process_data_events(time_limit=1)
            except pika.exceptions.ConnectionClosedByBroker as exc:
                LOGGER.info('Connection was closed by broker. Reason: "%s". Shutting down...', exc.reply_text)
            except pika.exceptions.ChannelClosedByBroker as exc:
                LOGGER.info('Channel was closed by broker. Reason: "%s". retrying...', exc.reply_text)
                self._connect_rabbitmq()
            except pika.exceptions.AMQPConnectionError:
                LOGGER.info("Connection was lost, retrying...")
                self._connect_rabbitmq()

    def _send_start_work_flow(self, job_id: uuid4, work_flow_type: WorkFlowType):
        # TODO convert to protobuf
        # TODO job_id converted to string for json
        body = json.dumps({"job_id": str(job_id)})
        self._send_output(Queue.from_workflow_type(work_flow_type), body)

    def _send_output(self, queue: Queue, message: str):
        body: bytes = message.encode("utf-8")
        self.rabbitmq_connection.add_callback_threadsafe(
            functools.partial(
                self.rabbitmq_channel.basic_publish, exchange=self.rabbitmq_exchange, routing_key=queue.value, body=body
            )
        )

    def _stop_rabbitmq(self):
        self.rabbitmq_is_running = False
        if self.rabbitmq_connection:
            self.rabbitmq_connection.add_callback_threadsafe(self.rabbitmq_connection.close)
